/*
 * System Interface Library for games
 * Copyright (c) 2007-2018 Andrew Church <achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/sysdep/windows/time.c: Timekeeping functions for Windows.
 */

#define IN_SYSDEP

#include "src/base.h"
#include "src/math.h"
#include "src/sysdep.h"
#include "src/sysdep/windows/internal.h"
#include "src/time.h"

/*************************************************************************/
/****************************** Local data *******************************/
/*************************************************************************/

/* Divisor for system-reported time (so the frequency reported by
 * sys_time_unit() fits in 31 bits). */
static uint8_t time_shift = 0;

/* Tick frequency returned from sys_time_unit().  Equal to the frequency
 * returned from QueryPerformanceFrequency() shifted right by time_shift
 * bits. */
static int ticks_per_sec;

/* Last value returned from QueryPerformanceCounter(). */
static uint64_t last_time;

/* Wraparound offset in case we're shifting the raw time values. */
static uint64_t time_offset = 0;

/*************************************************************************/
/************************** Interface routines ***************************/
/*************************************************************************/

void sys_time_init(void)
{
    LARGE_INTEGER frequency_buf;
    QueryPerformanceFrequency(&frequency_buf);
    uint64_t frequency = frequency_buf.QuadPart;
    time_shift = 0;
    while (frequency >= (uint64_t)1 << 31) {
        time_shift++;
        frequency >>= 1;
    }
    ticks_per_sec = (int)frequency;

    LARGE_INTEGER now_buf;
    QueryPerformanceCounter(&now_buf);
    last_time = now_buf.QuadPart;
}

/*-----------------------------------------------------------------------*/

int sys_time_unit(void)
{
    return ticks_per_sec;
}

/*-----------------------------------------------------------------------*/

uint64_t sys_time_now(void)
{
    LARGE_INTEGER now_buf;
    QueryPerformanceCounter(&now_buf);
    const uint64_t now = now_buf.QuadPart;
    if (UNLIKELY(now < last_time) && time_shift > 0) {
        time_offset += (uint64_t)1 << (64 - time_shift);
    }
    return (now >> time_shift) + time_offset;
}

/*-----------------------------------------------------------------------*/

void sys_time_delay(int64_t time)
{
    const double sec = (double)time / (double)ticks_per_sec;
    /* Round so that the value passed to Sleep() is not less than the
     * original request, but try to avoid incrementing the millisecond
     * count due to rounding error. */
    const int msec = iround((sec * 1000) + 0.001);
    Sleep(msec);
}

/*-----------------------------------------------------------------------*/

int sys_time_get_utc(struct DateTime *time_ret)
{
    SYSTEMTIME time;
    GetSystemTime(&time);
    time_ret->year = time.wYear;
    time_ret->month = time.wMonth;
    time_ret->day = time.wDay;
    time_ret->weekday = time.wDayOfWeek;
    time_ret->hour = time.wHour;
    time_ret->minute = time.wMinute;
    time_ret->second = time.wSecond;
    time_ret->nsec = time.wMilliseconds * 1000000;

    TIME_ZONE_INFORMATION tzi;
    switch (GetTimeZoneInformation(&tzi)) {
      case TIME_ZONE_ID_UNKNOWN:
        return -(tzi.Bias);
      case TIME_ZONE_ID_STANDARD:
        return -(tzi.Bias + tzi.StandardBias);
      case TIME_ZONE_ID_DAYLIGHT:
        return -(tzi.Bias + tzi.DaylightBias);
      default: {
        static uint8_t warned = 0;
        if (!warned) {
            DLOG("Failed to get time zone information: %s",
                 windows_strerror(GetLastError()));
            warned = 1;
        }
        return 0;
      }
    }
}

/*************************************************************************/
/*************************************************************************/
