/*
 * System Interface Library for games
 * Copyright (c) 2007-2018 Andrew Church <achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/test/sysdep/linux/movie.c: Tests for the FFmpeg-based Linux
 * implementation of the system-level movie playback interface.
 */

#include "src/base.h"
#include "src/graphics.h"
#include "src/math.h"
#include "src/movie.h"
#include "src/sound.h"
#include "src/sound/mixer.h"  // For sound_mixer_get_pcm().
#include "src/sysdep.h"
#include "src/sysdep/test.h"
#include "src/test/base.h"
#include "src/test/graphics/internal.h"
#include "src/test/movie/internal.h"
#include "src/time.h"

/*************************************************************************/
/****************************** Test runner ******************************/
/*************************************************************************/

static int do_test_linux_movie(void);

int test_linux_movie(void)
{
    /* If FFmpeg isn't enabled, there's nothing to test. */
#if !defined(SIL_PLATFORM_LINUX_USE_FFMPEG)
    SKIP("FFmepg not enabled.");
#endif

    return run_tests_in_window(do_test_linux_movie);
}

DEFINE_GENERIC_TEST_RUNNER(do_test_linux_movie)

/*-----------------------------------------------------------------------*/

TEST_INIT(init)
{
    if (strcmp(CURRENT_TEST_NAME(), "test_core") == 0) {
        return 1;
    }

    CHECK_TRUE(sys_file_init());
    /* Prime any statically-allocated low-level resources (e.g. ioqueue)
     * so CHECK_MEMORY_FAILURES doesn't report spurious leaks. */
    SysFile *fh;
    char path[4096+20];
    ASSERT(sys_get_resource_path_prefix(path, sizeof(path))
           < (int)sizeof(path));
    ASSERT(strformat_check(path+strlen(path), sizeof(path)-strlen(path),
                           "testdata/test.txt"));
    ASSERT(fh = sys_file_open(path));
    char buf[1];
    int req;
    ASSERT(req = sys_file_read_async(fh, buf, 1, 0, -1));
    ASSERT(sys_file_wait_async(req) == 1);
    sys_file_close(fh);

    sys_test_sound_set_output_rate(44100);
    sound_init();
    CHECK_TRUE(sound_open_device("", 2));

    time_init();

    graphics_start_frame();
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST_CLEANUP(cleanup)
{
    if (strcmp(CURRENT_TEST_NAME(), "test_core") == 0) {
        return 1;
    }

    graphics_finish_frame();
    graphics_flush_resources();
    sound_cleanup();
    sys_file_cleanup();
    return 1;
}

/*************************************************************************/
/***************************** Test routines *****************************/
/*************************************************************************/

TEST(test_core)
{
    /* These tests will only work if FFmpeg is enabled; they are only
     * meaningful if WebM software decoding is also enabled (so the core
     * tests use the software decoder). */
#if !defined(SIL_MOVIE_INCLUDE_WEBM)
    return 1;  // The sysdep code was already tested by the core.
#elif !defined(SIL_PLATFORM_LINUX_USE_FFMPEG)
    SKIP("FFmpeg not enabled.");
#else
    if (!test_movie_core_with_extension(".mp4")) {
        FAIL("Previous failure(s) occurred using sysdep/linux movie interface");
    }
    return 1;
#endif
}

/*-----------------------------------------------------------------------*/

TEST(test_pcm_s16le_audio)
{
    int movie;
    CHECK_TRUE(movie = movie_open("testdata/movie/test-pcm.avi", 0, 0));

    CHECK_TRUE(movie_play(movie));
    for (int frame = 0; frame < MOVIE_FRAMES; frame++) {
        graphics_clear(0, 0, 0, 0, 1, 0);
        if (!movie_next_frame(movie)) {
            FAIL("movie_next_frame() failed for frame %d", frame);
        }

        /* Unlike the tests in src/test/movie/core.c, these samples should
         * be exact modulo rounding error (since the file contains raw PCM). */
        int16_t pcm[MOVIE_SAMPLES_PER_FRAME*2];
        const int num_samples = lenof(pcm) / 2;
        sound_update();
        sound_mixer_get_pcm(pcm, num_samples);
        const int left_period = 100;
        const int left_amplitude = 10000;
        const int right_period = 50;
        const int right_amplitude = 6000;
        int audio_pos = frame * MOVIE_SAMPLES_PER_FRAME;
        for (int i = 0; i < num_samples; i++, audio_pos++) {
            const int left_phase = audio_pos % left_period;
            const int left_expected =
                iroundf(left_amplitude * dsinf(left_phase*360.0f/left_period));
            if (abs(pcm[i*2+0] - left_expected) > 1) {
                FAIL("Audio sample %d (left) was %d but should have been %d",
                     i, pcm[i*2+0], left_expected);
            }
            const int right_phase = audio_pos % right_period;
            const int right_expected =
                iroundf(right_amplitude
                        * dsinf(right_phase*360.0f/right_period));
            if (abs(pcm[i*2+1] - right_expected) > 1) {
                FAIL("Audio sample %d (right) was %d but should have been %d",
                     i, pcm[i*2+1], right_expected);
            }
        }

        graphics_finish_frame();
        graphics_start_frame();
    }

    movie_close(movie);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_no_video)
{
    CHECK_FALSE(movie_open("testdata/sound/square.ogg", 0, 0));
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_unsupported_pixel_format)
{
    CHECK_FALSE(movie_open("testdata/movie/test-yuv422.mp4", 0, 0));
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_too_many_audio_channels)
{
    CHECK_FALSE(movie_open("testdata/movie/test-surround.mp4", 0, 0));
    return 1;
}

/*************************************************************************/
/*************************************************************************/
